import numpy as np
from GRAIL.Correlation import Correlation
import heapq
import GRAIL.OPQ as OPQ
import GRAIL.PQ as PQ
from time import time

def kNN(TRAIN, TEST, method, k, representation=None, use_exact_rep = False,
        pq_method = None, Ks = 4, M = 16, **kwargs):
    """
    Approximate or exact k-nearest neighbors algorithm depending on the representation
    :param TRAIN: The training set to get the neighbors from
    :param TEST: The test set whose neighbors we get
    :param method: The correlation or distance measure being used
    :param k: how many nearest neighbors to return
    :param representation: The representation being used if any, for instance GRAIL. This is a representation object.
    :param **kwargs: arguments for the correlator
    :return: a matrix of size row(TEST)xk
    """
    if k > TRAIN.shape[0]:
        raise ValueError("Number of nearest neighbors should be smaller than number of points")

    if pq_method:
        return kNN_with_pq(TRAIN, TEST, method, k, representation, use_exact_rep, pq_method, Ks, M, **kwargs)

    rowTEST = TEST.shape[0]
    rowTRAIN = TRAIN.shape[0]

    neighbors = np.zeros((rowTEST, k))
    correlations = np.zeros((rowTEST, k))
    if representation:
        TRAIN, TEST = representation.get_rep_train_test(TRAIN, TEST, exact=use_exact_rep)

    t = time()
    for i in range(rowTEST):
        x = TEST[i, :]
        corr_array = np.zeros(rowTRAIN)
        for j in range(rowTRAIN):
            y = TRAIN[j, :]
            correlation = Correlation(x, y, correlation_protocol_name=method, **kwargs)
            corr_array[j] = correlation.correlate()
        if Correlation.is_similarity(method):
            temp = np.array(heapq.nlargest(k, enumerate(corr_array), key = lambda x: x[1]))
            neighbors[i,:] = temp[:, 0]
            correlations[i,:] = temp[:,1]
        else:
            temp = np.array(heapq.nsmallest(k, enumerate(corr_array), key = lambda x: x[1]))
            neighbors[i,:] = temp[:, 0]
            correlations[i,:] = temp[:,1]
    neighbors = neighbors.astype(int)
    return_time = time() - t
    return neighbors, correlations, return_time

#check the returned distances
def kNN_with_pq(TRAIN, TEST, method, k, representation=None, use_exact_rep = False,
                pq_method = "opq", Ks = 4, M = 16,**kwargs):
    if method != "ED":
        raise ValueError("Product Quantization can only be used with ED.")

    rowTEST = TEST.shape[0]
    rowTRAIN = TRAIN.shape[0]

    neighbors = np.zeros((rowTEST, k))
    distances = np.zeros((rowTEST, k))
    if representation:
        together = np.vstack((TRAIN, TEST))
        if use_exact_rep:
            rep_together = representation.get_exact_representation(together)
        else:
            rep_together = representation.get_representation(together)
        TRAIN = rep_together[0:rowTRAIN, :]
        TEST = rep_together[rowTRAIN:, :]


    if rowTRAIN < Ks:
        Ks = 2 ** int(np.floor(np.log2(rowTRAIN)))

    # This code trims the last parts
    # if TRAIN.shape[1] > M and TRAIN.shape[1] % M != 0:
    #     TRAIN = TRAIN[:, 0:(TRAIN.shape[1] - TRAIN.shape[1] % M)]
    #     TEST = TEST[:, 0:(TRAIN.shape[1] - TRAIN.shape[1] % M)]

    # padding with up to 2^n
    next_pow_2 = int(max(np.ceil(np.log2(TRAIN.shape[1])), np.ceil(np.log2(M))))
    TRAIN = np.hstack((TRAIN, np.zeros((rowTRAIN, 2 ** next_pow_2-TRAIN.shape[1]))))
    TEST = np.hstack((TEST, np.zeros((rowTEST, 2 ** next_pow_2 - TEST.shape[1]))))

    TRAIN = TRAIN.astype(np.float32)
    TEST = TEST.astype(np.float32)



    if pq_method == "opq":
        pq = OPQ.OPQ(M=M, Ks= Ks, verbose=False)
    elif pq_method == "pq":
        pq = PQ.PQ(M=M, Ks= Ks, verbose=False)
    else:
        raise ValueError("Product quantization method not found.")


    pq.fit(vecs=TRAIN)
    TRAIN_code = pq.encode(vecs=TRAIN)
    t = time()

    for i in range(rowTEST):
        query = TEST[i, :]
        dists = pq.dtable(query=query).adist(codes=TRAIN_code)
        temp = np.array(heapq.nsmallest(k, enumerate(dists), key = lambda x: x[1]))
        neighbors[i,:] = temp[:, 0]
        distances[i,:] = temp[:,1]
    neighbors = neighbors.astype(int)
    return_time = time() -t
    #print("Time for M = ", M , ": ", return_time)
    return neighbors, distances, return_time


def kNN_classifier(TRAIN, train_labels, TEST, method, k, representation=None, use_exact_rep = False,
                   pq_method = None, Ks = 4, M = 16, **kwargs):
    neighbors, _, return_time = kNN(TRAIN, TEST, method, k, representation, use_exact_rep, pq_method, Ks, M, **kwargs)
    return_labels = np.zeros(TEST.shape[0])
    for i in range(TEST.shape[0]):
        nearest_labels = np.zeros(k)
        for j in range(k):
            nearest_labels[j] = train_labels[neighbors[i,j]]
        unique, counts = np.unique(nearest_labels, return_counts=True)
        mx = 0
        mx_label = 0
        for j in range(unique.shape[0]):
            if counts[j] > mx:
                mx = counts[j]
                mx_label = unique[j]
        return_labels[i] = mx_label
    return return_labels, return_time

def kNN_classification_precision_test(exact_neighbors, TRAIN, train_labels, TEST, test_labels, method, k, representation=None, use_exact_rep = False,
                   pq_method = None, Ks = 4, M = 16, **kwargs):
    neighbors, _, return_time = kNN(TRAIN, TEST, method, k, representation,
                         use_exact_rep, pq_method, Ks, M, **kwargs)
    return_labels = np.zeros(TEST.shape[0])

    tp_arr = np.zeros(TEST.shape[0])

    for i in range(TEST.shape[0]):
        for j in range(k):
            if neighbors[i,j] in exact_neighbors[i]:
                tp_arr[i] = tp_arr[i] + 1

    precision = np.mean(tp_arr) / k

    #assign labels
    for i in range(TEST.shape[0]):
        nearest_labels = np.zeros(k)
        for j in range(k):
            nearest_labels[j] = train_labels[neighbors[i,j]]
        unique, counts = np.unique(nearest_labels, return_counts=True)
        mx = 0
        mx_label = 0
        for j in range(unique.shape[0]):
            if counts[j] > mx:
                mx = counts[j]
                mx_label = unique[j]
        return_labels[i] = mx_label

    # accuracy of the labels
    cnt_acc = 0

    for i in range(test_labels.shape[0]):
        if test_labels[i] == return_labels[i]:
            cnt_acc += 1
    classification_accuracy = cnt_acc / test_labels.shape[0]
    return classification_accuracy, precision, return_time

#Stands for mean average precision
def MAP(exact_neighbors, neighbors):
    n = exact_neighbors.shape[0]
    k = exact_neighbors.shape[1]
    map_measure = 0
    for i in range(n):
        AP = 0
        psum = 0
        for r in range(k):
            if neighbors[i,r] in exact_neighbors[i]:
                rel = 1
            else:
                rel = 0
            psum += rel
            AP += (psum / (r+1))*rel
        AP /= k
        map_measure += AP
    map_measure /= n
    return map_measure

def avg_recall_measure(exact_neighbors, neighbors):
    n = exact_neighbors.shape[0]
    k = exact_neighbors.shape[1]

    avg_recall = 0
    for i in range(n):
        recall = 0
        for r in range(k):
            if neighbors[i,r] in exact_neighbors[i]:
                recall += 1
        recall /= k
        avg_recall += recall
    avg_recall /= n
    return avg_recall



import numpy as np
import GRAIL.SINK as SINK
from scipy import stats
from scipy import linalg
import random
import math
import heapq


def zscore(x, axis=0, ddof=0):
    return np.nan_to_num(stats.zscore(x, axis=axis, ddof=ddof))


# somewhat tested
def sbd(x, y):
    '''
    Shape based distance
    :param x: z-normalized time series x
    :param y: z-normalized time series y
    :return: The distance, the shift, and the aligned sequence
    '''
    ncc_seq = SINK.NCC(x, y)
    index = np.argmax(ncc_seq)
    value = ncc_seq[index]

    dist = 1 - value
    shift = index - len(x) + 1  # make sure this is true
    if shift > 0:
        yshifted = np.concatenate((np.zeros(shift), y[0:-shift]))
    elif shift == 0:
        yshifted = y
    else:
        yshifted = np.concatenate((y[-shift:], np.zeros(-shift)))

    return [dist, shift, yshifted]


def kshape_centroid(X, mem, ref_seq, k):
    '''
    Computes the centroid for the kshape algorithm
    :param X: the matrix of time series
    :param mem: partition membership array
    :param ref_seq: the reference sequence time series are aligned against
    :param k: the partition number we want the centroid for
    :return: centroid of partition k
    '''
    partition = np.array([]).reshape(0, X.shape[1])
    for i in range(X.shape[0]):
        if mem[i] == k:
            partition = np.vstack((partition, X[i, :]))

    # return all zeros if partition is empty
    if partition.shape[0] == 0:
        return np.zeros((1, X.shape[1]))

    if sum(ref_seq) != 0:
        for i in range(partition.shape[0]):
            [dist, shift, vshifted] = sbd(ref_seq, partition[i, :])
            partition[i, :] = vshifted

    n = partition.shape[0]
    m = partition.shape[1]
    z_partition = zscore(partition, axis=1, ddof=1)

    S = np.transpose(z_partition) @ z_partition
    Q = np.identity(m) - (1 / m) * np.ones(m)
    M = np.transpose(Q) @ S @ Q
    eigval, centroid = linalg.eigh(M, subset_by_index=[m - 1, m - 1])
    centroid = centroid.transpose()

    d1 = ED(partition[0, :], centroid)
    d2 = ED(partition[0, :], -centroid)
    if d1 < d2:
        return centroid
    else:
        return -centroid


def ED(x, y):
    return np.sqrt(np.sum(np.power(x - y, 2)))


def matlab_kshape(A, k):
    '''
    shape based clustering algorithm
    This is the version where we randomly initialize the partitions
    :param X: mxn matrix containing time series that are z-normalized
    :param k: number of clusters
    :return: index is the length n array containing the index of the clusters to which
    the series are assigned. centroids is the kxm matrix containing the centroids of
    the clusters
    '''

    m = A.shape[0]
    mem = np.zeros(m)

    for i in range(m):
        mem[i] = random.randrange(k)
    cent = np.zeros((k, A.shape[1]))

    for iter in range(100):
        prev_mem = mem.copy()
        cluster_cnt = np.zeros(k)
        empty_cluster_cnt = 0
        D = math.inf * np.ones((m, k))

        for i in range(k):
            cent[i, :] = kshape_centroid(A, mem, cent[i, :], i)
            cent[i, :] = zscore(cent[i, :], ddof=1)

        for i in range(m):
            for j in range(k):
                dist = 1 - max(SINK.NCC(A[i, :], cent[j, :]))
                D[i, j] = dist

        for i in range(m):
            mem[i] = np.argmin(D[i, :])
            cluster_cnt[int(mem[i])] = cluster_cnt[int(mem[i])] + 1

        # check for empty clusters
        empty_cluster_list = []
        for cluster in range(k):
            if cluster_cnt[cluster] == 0:
                empty_cluster_cnt = empty_cluster_cnt + 1
                empty_cluster_list.append(cluster)

        # deal with empty clusters
        if empty_cluster_cnt != 0:
            min_dists = np.amin(D, axis=1)
            templist = np.array(heapq.nlargest(empty_cluster_cnt, enumerate(min_dists), key = lambda x: x[1]))
            distant_points = templist[:, 0]
            for i in range(empty_cluster_cnt):
                mem[int(distant_points[i])] = empty_cluster_list[i]


        if linalg.norm(prev_mem - mem) == 0:
            for i in range(k):
                cent[i, :] = kshape_centroid(A, mem, cent[i, :], i)
                cent[i, :] = zscore(cent[i, :], ddof=1)
            break

    return [mem, cent]


def kshape_with_centroid_initialize(X, k, is_pp = True):
    '''
    shape based clustering algorithm
    This is the version where we randomly initialize the centroids.
    :param X: nxm matrix containing time series that are z-normalized
    :param k: number of clusters
    :param is_pp: if true, use k-shape++ initialization method.
    :return: mem is the length n array containing the index of the clusters to which
    the series are assigned. centroids is the kxm matrix containing the centroids of
    the clusters
    '''

    # initialization
    n = X.shape[0]
    mem = np.zeros(n)
    if is_pp:
        centroids = kshape_pp_initialization(X,k)
    else:
        initial_centroids = random.sample(range(n), k)
        centroids = X[initial_centroids, :]

    for iter in range(100):
        print(iter)
        prev_mem = mem.copy()
        cluster_cnt = np.zeros(k)
        empty_cluster_cnt = 0
        D = math.inf * np.ones((n, k))
        # assignment
        for i in range(n):
            for j in range(k):
                dist = 1 - max(SINK.NCC(X[i, :], centroids[j, :]))
                D[i, j] = dist

        for i in range(n):
            mem[i] = np.argmin(D[i, :])
            cluster_cnt[int(mem[i])] = cluster_cnt[int(mem[i])] + 1

        if linalg.norm(mem - prev_mem) == 0:
            break

        # check for empty clusters
        empty_cluster_list = []
        for cluster in range(k):
            if cluster_cnt[cluster] == 0:
                empty_cluster_cnt = empty_cluster_cnt + 1
                empty_cluster_list.append(cluster)

        # deal with empty clusters
        if empty_cluster_cnt != 0:
            min_dists = np.amin(D, axis=1)
            templist = np.array(heapq.nlargest(empty_cluster_cnt, enumerate(min_dists), key = lambda x: x[1]))
            distant_points = templist[:, 0]
            for i in range(empty_cluster_cnt):
                mem[int(distant_points[i])] = empty_cluster_list[i]

        # refinement
        for i in range(k):
            centroids[i, :] = kshape_centroid(X, mem, centroids[i, :], i)
            centroids[i, :] = zscore(centroids[i, :])


    return [mem, centroids]


def kshape_pp_initialization(X, k):
    """
    This is based on the k-means++ algorithm. It is an initialization method designed
    to avoid bad initial clusters.
    :param X: Matrix of time series
    :param k: number of clusters
    :return: centroids
    """
    n = X.shape[0]
    centers = np.zeros((k,X.shape[1]))
    ind = random.randrange(k)
    centers[0,:] = X[ind, :]
    for i in range(1,k):
        D = np.ones((n,i)) * np.inf
        weights = np.zeros(n)
        for j in range(n):
            for c in range(i):
                D[j,c] = 1 - max(SINK.NCC(X[j,:], centers[c,:]))
            weights[j] = min(D[j,:])
        ind = random.choices(list(range(n)), weights=weights, k=1)
        centers[i,:] = X[ind,:]
    return centers



import numpy as np


def SINK(x, y, gamma_val, k=-1, e=-1):
    check_arguments(k, e)
    return sumExpNCC(x, y, gamma_val, k, e) / np.sqrt(sumExpNCC(x, x, gamma_val, k, e) * sumExpNCC(y, y, gamma_val, k, e))


# Helper function for SINK
def sumExpNCC(x, y, gamma, k=-1, e=-1):
    check_arguments(k, e)
    return np.sum(np.exp(gamma * NCC(x, y, k, e)))


def NCC(x, y, k=-1, e=-1):
    '''
    Normalized cross correlation
    :param x: time series
    :param y: time series
    :param k: optional parameter for compression
    :param e: preserved energy in the Fourier Domain
    :return: cross correlation sequence
    '''

    check_arguments(k, e)
    length = max(len(x), len(y))

    fftlength = int(pow(2, nextPow2(2 * length - 1)))

    if k == -1 and e == -1:
        fftx = np.fft.fft(x, n=fftlength)
        ffty = np.fft.fft(y, n=fftlength)
    elif k == -1 and e != -1:
        fftx = preserved_energy(x, e, fftlength)
        ffty = preserved_energy(x, e, fftlength)
    else:
        fftx = leadingFourier(np.fft.fft(x, n=fftlength), k)
        ffty = leadingFourier(np.fft.fft(y, n=fftlength), k)
    r = np.fft.ifft(np.multiply(fftx, np.conj(ffty)))

    end = len(r) - 1
    r = np.concatenate((r[(end - length + 2):len(r)], r[0: length]))

    result = np.divide(np.real(r), (np.linalg.norm(x) * np.linalg.norm(y)))
    return np.nan_to_num(result)


# Helper function for NCC
def nextPow2(x):
    return np.ceil(np.log2(abs(x)))


def preserved_energy(x, e, fftlength):
    fftx = np.fft.fft(x, n=fftlength)
    norm_cumsum = np.divide(np.cumsum(np.power(np.abs(fftx), 2)), np.sum(np.power(np.abs(fftx), 2)))
    k = np.argwhere(norm_cumsum >= e / 2)[0, 0] + 1
    return leadingFourier(x, k)


# Helper function for NCC
def leadingFourier(x, k):
    m = int(np.floor(len(x) / 2) + 1)
    x[k: (m - 1 + m - k)] = 0
    return x


def check_arguments(k, e):
    if k != -1 and e != -1:
        raise ValueError(
            "Preserved energy and number of Fourier coefficients should not be specified at the same time.")



from GRAIL.SINK import SINK
import numpy as np
import time
from GRAIL.FrequentDirections import FrequentDirections
from GRAIL.kshape import matlab_kshape, kshape_with_centroid_initialize
import GRAIL.exceptions as exceptions

def approx_gte(x, y):
    return np.logical_or((x > y), np.isclose(x, y))


def GRAIL_general(X, kernel, d, f, r, kernel_param_range, eigenvecMatrix = None,
              inVa = None, initialization_method = "partition", dictionary_sampling = "kshape", **kwargs):
    gamma = None
    n = X.shape[0]

    #Dictionary Sampling
    if dictionary_sampling == "kshape":
        if initialization_method == "partition":
            [mem, Dictionary] = matlab_kshape(X,d)
        elif initialization_method == "centroid_uniform":
            [mem, Dictionary] = kshape_with_centroid_initialize(X, d, is_pp=False)
        elif initialization_method == "k-shape++":
            [mem, Dictionary] = kshape_with_centroid_initialize(X, d, is_pp=True)
        else:
            raise exceptions.InitializationMethodNotFound

    elif dictionary_sampling == "random":
        dict_indices = np.random.choice(n, d, replace = False)
        Dictionary = X[dict_indices, :]

    else:
        raise exceptions.SamplingMethodNotFound


    # Tune the parameter of the kernel
    if kwargs['gamma_val'] == None and kernel == SINK:
        [score, kwargs['gamma_val']] = hyperparameter_select(Dictionary, kernel_param_range, r, kernel)
        gamma = kwargs['gamma_val']

    E = np.zeros((n, d))

    for i in range(n):
        for j in range(d):
            E[i, j] = kernel(X[i, :], Dictionary[j, :], **kwargs)

    if eigenvecMatrix == None and inVa == None:
        W = np.zeros((d, d))
        for i in range(d):
            for j in range(d):
                W[i, j] = kernel(Dictionary[i, :], Dictionary[j, :], **kwargs)

        [eigenvalvector, eigenvecMatrix] = np.linalg.eigh(W)
        inVa = np.diag(np.power(eigenvalvector, -0.5))

    Zexact = E @ eigenvecMatrix @ inVa
    Zexact = CheckNaNInfComplex(Zexact)
    Zexact = np.real(Zexact)

    BSketch = fd(Zexact, int(np.ceil(0.5 * d)))

    # eigh returns sorted eigenvalues in ascending order. We reverse this.
    [eigvalues, Q] = np.linalg.eigh(np.matrix.transpose(BSketch) @ BSketch)
    eigvalues = np.real(eigvalues)
    Q = np.real(Q)
    eigvalues = np.flip(eigvalues)
    Q = np.flip(Q)

    VarExplainedCumSum = np.divide(np.cumsum(eigvalues), np.sum(eigvalues))
    k = np.argwhere(approx_gte(VarExplainedCumSum, f))[0, 0] + 1
    Z_k = CheckNaNInfComplex(Zexact @ Q[0:d, 0:k])
    return Z_k, Zexact, gamma


def GRAIL_rep(X, d, f, r, GV, fourier_coeff = -1, e = -1, eigenvecMatrix = None, inVa = None, gamma = None, initialization_method = "partition"):
    """
    :param X: nxm matrix that contains n time series
    :param d: number of landmark series to extract from kshape
    :param f: scalar to tune the dimensionality k of Z_k
    :param r: parameter for tuning gamma, taken as 20 in the paper
    :param GV: vector of gammas to select the best gamma
    :param fourier_coeff: number of fourier coeffs to keep
    :param e: preserved energy in Fourier domain
    :return: Z_k nxk matrix of low dimensional reduced representation
    """

    n = X.shape[0]
    if initialization_method == "partition":
        [mem, Dictionary] = matlab_kshape(X,d)
    elif initialization_method == "centroid_uniform":
        [mem, Dictionary] = kshape_with_centroid_initialize(X, d, is_pp=False)
    elif initialization_method == "k-shape++":
        [mem, Dictionary] = kshape_with_centroid_initialize(X, d, is_pp=True)
    else:
        raise exceptions.InitializationMethodNotFound

    if gamma == None:
        [score, gamma] = gamma_select(Dictionary, GV, r)

    E = np.zeros((n, d))
    for i in range(n):
        for j in range(d):
            E[i, j] = SINK(X[i, :], Dictionary[j, :], gamma,fourier_coeff, e)

    if eigenvecMatrix == None and inVa == None:
        W = np.zeros((d, d))
        for i in range(d):
            for j in range(d):
                W[i, j] = SINK(Dictionary[i, :], Dictionary[j, :], gamma,fourier_coeff, e)

        [eigenvalvector, eigenvecMatrix] = np.linalg.eigh(W)
        inVa = np.diag(np.power(eigenvalvector, -0.5))

    Zexact = E @ eigenvecMatrix @ inVa
    Zexact = CheckNaNInfComplex(Zexact)
    Zexact = np.real(Zexact)

    BSketch = fd(Zexact, int(np.ceil(0.5 * d)))

    # eigh returns sorted eigenvalues in ascending order. We reverse this.
    [eigvalues, Q] = np.linalg.eigh(np.matrix.transpose(BSketch) @ BSketch)
    eigvalues = np.real(eigvalues)
    Q = np.real(Q)
    eigvalues = np.flip(eigvalues)
    Q = np.flip(Q)

    VarExplainedCumSum = np.divide(np.cumsum(eigvalues), np.sum(eigvalues))
    k = np.argwhere(approx_gte(VarExplainedCumSum, f))[0, 0] + 1
    Z_k = CheckNaNInfComplex(Zexact @ Q[0:d, 0:k])
    return Z_k, Zexact, gamma


def CheckNaNInfComplex(Z):
    for i in range(Z.shape[0]):
        for j in range(Z.shape[1]):
            if np.isnan(Z[i, j]) or np.isinf(Z[i, j]) or (not np.isreal(Z[i, j])):
                Z[i, j] = 0

    return Z


#A function in the original MATLAB implementation. Not used here.
def repLearnKM(KM):
    t = time.time()
    [eigenvalvector, Q] = np.linalg.eigh(KM)
    eigenvalvector = np.flip(eigenvalvector)
    Q = np.flip(Q)

    VarExplainedCumSum = np.divide(np.cumsum(eigenvalvector), np.sum(eigenvalvector))

    DimFor99 = np.argwhere(VarExplainedCumSum >= 0.99)[0, 0] + 1
    DimFor98 = np.argwhere(VarExplainedCumSum >= 0.98)[0, 0] + 1
    DimFor97 = np.argwhere(VarExplainedCumSum >= 0.97)[0, 0] + 1
    DimFor95 = np.argwhere(VarExplainedCumSum >= 0.95)[0, 0] + 1
    DimFor90 = np.argwhere(VarExplainedCumSum >= 0.90)[0, 0] + 1
    DimFor85 = np.argwhere(VarExplainedCumSum >= 0.85)[0, 0] + 1
    DimFor80 = np.argwhere(VarExplainedCumSum >= 0.80)[0, 0] + 1

    RepLearnTime = time.time() - t

    Z99per = CheckNaNInfComplex(Q[:, 0: DimFor99] @ np.sqrt(np.diag(eigenvalvector[0: DimFor99])))
    Z98per = CheckNaNInfComplex(Q[:, 0: DimFor98] @ np.sqrt(np.diag(eigenvalvector[0: DimFor98])))
    Z97per = CheckNaNInfComplex(Q[:, 0: DimFor97] @ np.sqrt(np.diag(eigenvalvector[0: DimFor97])))
    Z95per = CheckNaNInfComplex(Q[:, 0: DimFor95] @ np.sqrt(np.diag(eigenvalvector[0: DimFor95])))
    Z90per = CheckNaNInfComplex(Q[:, 0: DimFor90] @ np.sqrt(np.diag(eigenvalvector[0: DimFor90])))
    Z85per = CheckNaNInfComplex(Q[:, 0: DimFor85] @ np.sqrt(np.diag(eigenvalvector[0: DimFor85])))
    Z80per = CheckNaNInfComplex(Q[:, 0: DimFor80] @ np.sqrt(np.diag(eigenvalvector[0: DimFor80])))

    Ztop20 = CheckNaNInfComplex(Q[:, 0: 20] @ np.sqrt(np.diag(eigenvalvector[0: 20])))
    Ztop10 = CheckNaNInfComplex(Q[:, 0: 10] @ np.sqrt(np.diag(eigenvalvector[0: 10])))
    Ztop5 = CheckNaNInfComplex(Q[:, 0: 5] @ np.sqrt(np.diag(eigenvalvector[0: 5])))
    return [Z99per, Z98per, Z97per, Z95per, Z90per, Z85per, Z80per, Ztop20, Ztop10, Ztop5, RepLearnTime]


def gamma_select(Dictionary, GV, r, k=-1):
    """
    Parameter Tuning function. Tunes the parameters for GRAIL
    This function does not work at the moment for gamma values that don't correspond
    to a range(n)
    :param Dictionary: Dictionary to summarize the dataset. Provided by KShape
    :param GV: A vector of Gamma values to choose from.
    :param k: Number of Fourier coeffs to keep when computing the SINK function.
    :param r: The number of top eigenvalues to be considered. This is 20 in the paper.
    :return: the tuned parameter gamma and its score
    """
    d = Dictionary.shape[0]
    GVar = np.zeros(len(GV) + 5)
    var_top_r = np.zeros(len(GV) + 5)
    score = np.zeros(len(GV) + 5)
    for gamma in GV:
        W = np.zeros((d, d))
        for i in range(d):
            for j in range(d):
                W[i, j] = SINK(Dictionary[i, :], Dictionary[j, :], gamma, k)
        GVar[gamma] = np.var(W)
        [eigenvalvector, eigenvecMatrix] = np.linalg.eigh(W)
        eigenvalvector = np.flip(eigenvalvector)
        eigenvecMatrix = np.flip(eigenvecMatrix)
        var_top_r[gamma] = np.sum(eigenvalvector[0:r]) / np.sum(eigenvalvector)

    score = GVar * var_top_r
    best_gamma = np.argmax(score)
    best_score = score[best_gamma]
    return [best_score, best_gamma]

def hyperparameter_select(Dictionary, GV, r, kernel):
    """
    Parameter Tuning function. Tunes the parameters for kernels
    :param Dictionary: Dictionary to summarize the dataset.
    :param GV: A vector of sigma values to choose from.
    :param r: The number of top eigenvalues to be considered. This is 20 in the paper.
    :return: the tuned parameter sigma and its score
    """
    d = Dictionary.shape[0]
    GVar = np.zeros(len(GV))
    var_top_r = np.zeros(len(GV))
    score = np.zeros(len(GV))
    for i in range(len(GV)):
        param = GV[i]
        W = np.zeros((d, d))
        for j in range(d):
            for k in range(d):
                W[j, k] = kernel(Dictionary[j, :], Dictionary[k, :], param)
        GVar[i] = np.var(W)
        [eigenvalvector, eigenvecMatrix] = np.linalg.eigh(W)
        eigenvalvector = np.flip(eigenvalvector)
        eigenvecMatrix = np.flip(eigenvecMatrix)
        var_top_r[i] = np.sum(eigenvalvector[0:r]) / np.sum(eigenvalvector)

    score = GVar * var_top_r
    best_param_index = np.argmax(score)
    best_param = GV[best_param_index]
    best_score = score[best_param_index]
    print("Selected parameter = ", best_param)
    return [best_score, best_param]


# Frequent directions helper function returns sketch matrix of size (ell x d)
def fd(A, ell):
    """
    Returns a sketch matrix of size ell x A.shape[1]
    :param A: Matrix to be sketched
    :param ell:
    :return:
    """
    d = A.shape[1]
    sketcher = FrequentDirections(d, ell)
    for i in range(A.shape[0]):
        sketcher.append(A[i, :])
    sketch = sketcher.get()
    return sketch


def choose_d(TRAIN, train_labels, TEST):
    return int(min(max(4*len(np.unique(train_labels)), np.ceil(0.4 * (TRAIN.shape[0] + TEST.shape[0])), 20), 100))




import numpy as np

#Gaussian Kernel
#tested
def DM2KM(DM):
    '''

    :param DM: nxn distance matrix of n time series
    :return: Kernel matrix
    '''

    n = DM.shape[0]
    sigma = np.mean(DM)
    KM = np.zeros((n,n))
    for i in range(n):
        for j in range(n):
            KM[i,j] = np.exp((-DM[i,j] ** 2)/(2*(sigma**2)))

    return KM

